package com.blackledge.david.tivo.game;

import java.awt.Color;
import java.awt.Rectangle;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import com.blackledge.david.game.BoardState;
import com.blackledge.david.game.Chat;
import com.blackledge.david.game.GameException;
import com.blackledge.david.game.MultiplayerGame;
import com.blackledge.david.game.MultiplayerGame.GameData;
import com.blackledge.david.game.PlayerState;
import com.blackledge.david.game.MultiplayerGame.GameInfo;

import com.blackledge.david.tivo.TiVoModelInfo;
import com.blackledge.david.tivo.V49BApplication;
import com.blackledge.david.tivo.V49BApplicationFactory;

import com.tivo.hme.bananas.BApplication;
import com.tivo.hme.bananas.BList;
import com.tivo.hme.bananas.BScreen;
import com.tivo.hme.bananas.BView;
import com.tivo.hme.sdk.HmeEvent;
import com.tivo.hme.sdk.Resolution;

public abstract class Game extends V49BApplication {
	public static abstract class GameFactory extends V49BApplicationFactory 
	{

		public int getBoardHeight() {
			return getGame().getBoardHeight();
		}

		public int getBoardWidth() {
			return getGame().getBoardWidth();
		}

		/** Instantiate game data, and return id for that instantiation */
		public synchronized int newGame() {
			return getGame().newGame();
		}
		
		public abstract MultiplayerGame getGame();

		/**
		 * Generic representation of a player doing something. It may or may not
		 * be successful depending on the rules of the game. A successful action
		 * may or may not affect the game state.
		 * 
		 * @param gameID
		 * @param playerIdentification
		 * @param action
		 * @param details
		 * @return true if the action was successful, false if it was not
		 *         allowed.
		 */
		public boolean playerAction(int gameID,
				PlayerState playerIdentification, int action, Object details) {
			return getGame().playerAction(gameID, playerIdentification, action, details);
		}
		
//		/**
//		 * Initialize a new game before players are added.  Mainly set game.boardState to a new instance
//		 * @param game
//		 */
//		protected abstract void initNewGame(GameData game);

		/**
		 * Add a player to a game and update state (e.g. start game) as appropriate.
		 * Throw full or id_in_use exceptions, modify player state or game as
		 * needed, and add player to game.players. Default impl throws
		 * ID_IN_USE if player is already in players, GAME_FULL if players already has maximumPlayers, otherwise adds player to
		 * game.players and if this player makes game meet minimumPlayers, FIRST player gets turn.
		 * Also chats "ENTERED"
		 * @param game
		 * @param player
		 * @throws GameException
		 */
		protected void addPlayer(GameData game, PlayerState player) throws GameException {
			// check the game isn't full of players already
			if(game.players.size() >= game.maximumPlayers) {
				throw GameException.GAME_FULL;
			}
			
			// check the playerid isn't already in the game
			if(game.players.contains(player))
				throw GameException.ID_IN_USE;

			game.players.add(player);
			
			playerChat(game.gameID, player, "ENTERED");
			
			if(game.players.size() == game.minimumPlayers) {
				//TODO if player 0 already HAS turn.. (leftover from abandonment) should somebody else get it to make sure turns stay balanced?
				game.players.get(0).hasTurn = true;
			}
		}

		/**
		 * Remove a player from a game and update state (e.g. assign remaining
		 * player as winner) as appropriate. remove player from game.players.
		 * Default Impl just removes player from game.players, first runs {@link #nextTurn(GameData, PlayerState)} if this player hasTurn.
		 * Also chat's "EXITED"
		 * @param game
		 * @param player
		 */
		protected void removePlayer(GameData game, PlayerState player) {
			// if player with turn is leaving, give turn to somebody remaining
			if(player.hasTurn) {
				nextTurn(game, player);
			}
			game.players.remove(player);
			
			playerChat(game.gameID, player, "EXITED");
		}

		/**
		 * change entries to move turn from player to next entry in game.players.
		 * @param game
		 * @param player
		 */
		protected void nextTurn(GameData game, PlayerState player) {
			int p1index = game.players.indexOf(player);
			int p2index;
			if(game.players.size() > p1index+1) {
				p2index = p1index+1;
			} else {
				p2index = 0;
			}
			player.hasTurn = false;
			game.players.get(p1index).hasTurn = false;
			
			game.players.get(p2index).hasTurn = true;
		}
		
		/**
		 * Get the FIRST player that has turn.
		 * @param game
		 * @return
		 */
		protected PlayerState getPlayerWithTurn(GameData game) {
			for(PlayerState player : game.players) {
				if(player.hasTurn) {
					return player;
				}
			}
			return null;
		}
		
		protected List<PlayerState> getPlayersWithTurn(GameData game) {
			ArrayList<PlayerState> result = new ArrayList<PlayerState>();
			for(PlayerState player : game.players) {
				if(player.hasTurn) {
					result.add(player);
				}
			}
			return result;
		}
		
		public void playerChat(int gameID, PlayerState speaker, String message) {
			Date timestamp = new Date();
			GameData gameData = getGame().getGameData(gameID);
			gameData.chat.add(new Chat(speaker, timestamp, message));
			gameData.lastChanged = timestamp;
			gameData.lastChangedChat = timestamp;
		}

		public void playerUpdate(int gameID, PlayerState player) {
			Date timestamp = new Date();
			GameData gameData = getGame().getGameData(gameID);
			int idx = gameData.players.indexOf(player);
			gameData.players.set(idx, new PlayerState(player));
			gameData.lastChanged = timestamp;
			gameData.lastChangedPlayers = timestamp;
		}

		private List<PlayerState> _connectedPlayers = new ArrayList<PlayerState>();
		public List<PlayerState> getConnectedPlayers() {
			synchronized (_connectedPlayers) {
				return new ArrayList<PlayerState>(_connectedPlayers);
			}
		}
		public List<PlayerState> getIdlePlayers() {
			List<PlayerState> players = getConnectedPlayers();
			for(GameInfo game : getGame().getGameList()) {
				players.removeAll(game.players);
			}
			return players;
		}

		public void connectPlayer(PlayerState player) {
			synchronized (_connectedPlayers) {
				if (!_connectedPlayers.contains(player)) {
					_connectedPlayers.add(player);
				}
			}
		}

		public void disconnectPlayer(PlayerState player) {
			synchronized (_connectedPlayers) {
				_connectedPlayers.remove(player);
			}
		}

	}
	

	
	
	protected PlayerState _player;
	public PlayerState getPlayer() {
		return _player;
	}
	
	protected int _gameID;

	public int getGameID() {
		return _gameID;
	}
	public void setGameID(int gameID) {
		_gameID = gameID;
	}
	

	/**
	 * Override this and call this implementation first (which sets app._player
	 * based on this.tivoID and connects the user to the game server) then you
	 * must push a custom instance of GameLobby
	 */
	@Override
	protected void showFirstPage() {
		TiVoModelInfo model = TiVoModelInfo.getInfo(this.tivoID);
		String name = (model == null) ? this.tivoID : model.getSeriesName() + "-"
				+ this.tivoID.substring(this.tivoID.length() - 4);
		_player = new PlayerState(this.tivoID, name);
		GameFactory gameFactory = (GameFactory) getFactory();
		gameFactory.connectPlayer(_player);
		
		push(newGameLobby(this), TRANSITION_FORWARD);
	}
	
	@Override
	protected void destroyCleanup() {
		GameFactory gameFactory = (GameFactory) getFactory();
		try {
			gameFactory.getGame().leaveGame(_gameID, _player);
		} catch (GameException e) {
			if(e.equals(GameException.GAME_NOT_FOUND)) {
				// expected
			} else {
			// TODO Auto-generated catch block
			e.printStackTrace();
			}
		}
		gameFactory.disconnectPlayer(_player);

		super.destroyCleanup();
	}

	private boolean _stopLobbyUpdate;
	protected void startLobbyUpdates() {
		updateLobbyNow();
		_stopLobbyUpdate = false;
		scheduleLobbyUpdate(0, 1000);
	}
	protected void stopLobbyUpdates() {
		_stopLobbyUpdate = true;
	}
	protected void updateLobbyNow() {
		processLobbyUpdate(0);
	}

	private boolean _stopPoll;
	protected void startGameUpdates() {
		// get starting game state
		updateGameNow();
		// schedule the first update
		_stopPoll = false;
		schedulePoll(_gameID, 1000);
	}
	protected void stopGameUpdates() {
		_stopPoll = true;
	}

	protected void updateGameNow() {
		processPoll(_gameID);
	}
	

	private Date _lastUpdated;

	
	
	private static final int POLL_KEY = KEY_LIVETV;
	private static final int POLL_TIME_DEFAULT = 500;

	private static final int LOBBY_UPDATE_KEY = KEY_TIVO;
	private static final int LOBBY_UPDATE_TIME_DEFAULT = 2000;
	
	@Override
	public boolean handleEvent(HmeEvent event) {
		switch (event.getOpCode()) {
		case EVT_KEY:
			HmeEvent.Key e = (HmeEvent.Key)event;
			if (e.getAction() == KEY_REPEAT) {
				switch (e.getCode()) {
				case POLL_KEY:
				{
					int arg = (int) e.getRawCode();
					int pollTime = processPoll(arg);
					//...repeat
					if(_stopPoll) {
						// when next we start, start from scratch.
						_lastUpdated = null;
//						_stopPoll = false;
						return true;
					}
					if(pollTime <= 0) {
						pollTime = POLL_TIME_DEFAULT;
					}
//					schedulePoll(arg, pollTime);
					schedulePoll(_gameID, pollTime);
					return true;
				}
				case LOBBY_UPDATE_KEY: {
					int arg = (int) e.getRawCode();
					int lobbyUpdateTime = processLobbyUpdate(arg);
					//...repeat
					if(_stopLobbyUpdate) {
//						_stopLobbyUpdate = false;
						return true;
					}
					if(lobbyUpdateTime <= 0) {
						lobbyUpdateTime = LOBBY_UPDATE_TIME_DEFAULT;
					}
					scheduleLobbyUpdate(arg, lobbyUpdateTime);
					return true;
				}
				}
			}
			break;
		}
		return super.handleEvent(event);
	}

	private void schedulePoll(int arg, int pollTime) {
		sendEvent(new HmeEvent.Key(getApp().getID(), KEY_REPEAT, POLL_KEY, arg), getResource("*"+pollTime));
	}
	
	private int processPoll(int arg) {
		if(getCurrentScreen() != null && getCurrentScreen() instanceof GameScreen && _gameID >= 0) {
			GameScreen screen = (GameScreen) getCurrentScreen();
			GameFactory gameFactory = (GameFactory) getFactory();
			GameData game = gameFactory.getGame().getGameData(_gameID);
			if (game != null && (_lastUpdated == null
					|| gameFactory.getGame().getGameUpdated(_gameID).after(_lastUpdated))) {
	
				if(_lastUpdated == null || gameFactory.getGame().getPlayersUpdated(_gameID).after(_lastUpdated)) {
					screen.updateGamePlayers(game.players);
				}
				if(_lastUpdated == null || gameFactory.getGame().getChatUpdated(_gameID).after(_lastUpdated)) {
					screen.updateGameChat(game.chat);
				}
				if(_lastUpdated == null || gameFactory.getGame().getBoardUpdated(_gameID).after(_lastUpdated)) {
					screen.updateGameBoard(game.boardState);
				}
				
				_lastUpdated = new Date();
			}
//		else
//			_lastUpdated = new Date();
		}
		
		return 0;
	}

	private void scheduleLobbyUpdate(int arg, int pollTime) {
		sendEvent(new HmeEvent.Key(getApp().getID(), KEY_REPEAT, LOBBY_UPDATE_KEY, arg), getResource("*"+pollTime));
	}
	
	private int processLobbyUpdate(int arg) {
		if(getCurrentScreen() != null && getCurrentScreen() instanceof GameLobby) {
			GameLobby screen = (GameLobby) getCurrentScreen();
			screen.setupList();
		}
		
		return 0;
	}
	
	public static abstract class GameScreen extends BScreen {

		private Object _chatFont = "default-15.font";
		private Color _chatColor = Color.white;
		private BView _chatText;
		protected BView _chatArea;
		private BView _playerArea;
		private HashMap<PlayerState, BView> _playerViews;
		private BView _notYourTurnIndicator;
		
		public GameScreen(BApplication app) {
			super(app);
			Rectangle titleArea = ((Game)app).getTitleRegion();
			Rectangle safeArea = ((Game)app).getSafeActionRegion();
			int chatWidth = titleArea.width*1/3;
			
			// create this (invisible) before chat/player views so it appears below them
			_notYourTurnIndicator = new BView(getAbove(), 0, 0, getWidth(), getHeight(),false);
			_notYourTurnIndicator.setResource(Color.black);
			_notYourTurnIndicator.setTransparency(0.5f);
			
			_chatArea = new BView(getAbove(), titleArea.x + titleArea.width-chatWidth, safeArea.y, chatWidth, safeArea.height);
			_chatArea.setResource(new Color(0,0,50));
			_chatText = new BView(_chatArea, 0, 0, _chatArea.getWidth(), _chatArea.getHeight());
			
			_playerViews = new HashMap<PlayerState, BView>();
		}


		/**
		 * Arg is expected to be the game id which will be passed to
		 * {@link GameFactory#joinGame(int, PlayerState)}
		 */
		@Override
		public boolean handleEnter(Object arg, boolean isReturn) {
			int chatFontSize = (int) (13 * ((Game)getBApp()).getHeightMultiplier());
			_chatFont = "default-" + chatFontSize + ".font";

//			System.out.println("ENTER with "+arg);
			if(arg != null && arg instanceof Integer) {
				Integer gameId = (Integer) arg;
				GameFactory gameFactory = (GameFactory) getBApp().getFactory();
				try {
					((Game)getBApp()).setGameID( gameFactory.getGame().joinGame(gameId, ((Game)getBApp())._player));
				} catch (GameException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				
				// build player area now that we know how many players the game is allowed to have.
				if(_playerArea == null) {
//					Rectangle titleArea = ((Game)getBApp()).getTitleRegion();
					int id = ((Game)getBApp()).getGameID();
//					_playerArea = new BView(getAbove(), titleArea.x, titleArea.y, titleArea.width/2, 20*gameFactory.games.get(id).maximumPlayers);
					int playerAreaHeight = 20*gameFactory.getGame().getGameData(id).maximumPlayers;
					_playerArea = new BView(getAbove(), 
							_chatArea.getX(), _chatArea.getY()+_chatArea.getHeight()-playerAreaHeight, 
							_chatArea.getWidth(), playerAreaHeight);
					_playerArea.setResource(Color.gray);
					// adjust chatarea to allow for playerarea
					_chatArea.setSize(_chatArea.getWidth(), _chatArea.getHeight() - _playerArea.getHeight());
					_chatText.setSize(_chatArea.getWidth(), _chatArea.getHeight());
				}
			}
			
			((Game)getBApp()).startGameUpdates();
			
			return super.handleEnter(arg, isReturn);
		}

		/**
		 * The area available for the game board, excluding the title, chat,
		 * player and any other decorative areas
		 */
		public Rectangle getGameRegion() {
			// using scroll region less space on the right for chat area
			Rectangle scrollRegion = ((Game) getBApp()).getScrollRegion();
			return new Rectangle(scrollRegion.x, scrollRegion.y,
					(_chatArea.getX() - scrollRegion.x),
					scrollRegion.height);
		}
		
		/**
		 * The area available for the game title, excluding the chat and any
		 * other decorative areas.  Excludes {@link Game#getLogoRegion()}
		 */
		protected Rectangle getGameTitleRegion() {
			// using title region less space on the right for chat area
			Rectangle titleRegion = ((Game) getBApp()).getTitleRegion();
			return new Rectangle(titleRegion.x, titleRegion.y,
					titleRegion.width - (_chatArea.getX() - titleRegion.x),
					titleRegion.height);
		}

		@Override
		public boolean handleExit() {
			((Game)getBApp()).stopGameUpdates();

			return super.handleExit();
		}
		
		@Override
		public boolean handleKeyPress(int code, long rawcode) {
			switch(code) {
			case KEY_CLEAR:
				Game game = (Game)getBApp();
				game.stopGameUpdates();
				GameFactory gameFactory = (GameFactory) game.getFactory();
				try {
					gameFactory.getGame().leaveGame(game.getGameID(), game._player);
				} catch (GameException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				((Game)getBApp()).setGameID(-1);
				getBApp().pop();
				return true;
			}
			return super.handleKeyPress(code, rawcode);
		}

		protected abstract void updateGameBoard(BoardState gameBoard);

		protected void updateGameChat(ArrayList<Chat> gameChat) {
			StringBuffer latestChats = new StringBuffer();
			for(int i = gameChat.size()-1 ; i >= 0 && i > gameChat.size()-50 ; --i) {
				latestChats.insert(0,"\n");
				latestChats.insert(0, gameChat.get(i));
			}
			
			if (_chatText != null) {
				if (_chatText.getResource() != null) {
					_chatText.getResource().remove();
				}
				_chatText.setResource(
						_chatText.createText(_chatFont, _chatColor,
								latestChats.toString()), 
								RSRC_HALIGN_LEFT | RSRC_VALIGN_BOTTOM | RSRC_TEXT_WRAP);
			}
		}

		protected void updateGamePlayers(ArrayList<PlayerState> gamePlayers) {
			if(_playerViews != null) {
				Iterator<BView> vals = _playerViews.values().iterator();
				while( vals.hasNext()) {
					BView view = vals.next();
					view.setVisible(false);
				}
				for(int i = 0 ; i < gamePlayers.size(); ++i) {
					PlayerState player = gamePlayers.get(i);
					BView view = _playerViews.get(player);
					// TODO extract old key to compare against new one to decide if update needed, and if so replace key
					if(view == null) {
						view = new BView(_playerArea, 0, 0, _playerArea.getWidth(), 20);
						new BView(view, 0, 0, view.getWidth(), view.getHeight());
						_playerViews.put(player, view);
					}
					view.setVisible(true);
					view.setLocation(0, 20*i, getResource("*500"));
					view.setResource(player.color);
					BView txt = (BView) view.getChild(0);
					if(txt.getResource() != null) {
						txt.getResource().remove();
					}
					String playertext = player.name+" ("+player.score+")";
					if(player.equals(((Game)getBApp())._player)) {
						// update player details
						((Game)getBApp())._player = new PlayerState(player);
						
						if(player.hasTurn) {
							_notYourTurnIndicator.setVisible(false);
							txt.setResource(txt.createText("default-18-bold.font", Color.white, playertext), RSRC_HALIGN_RIGHT|RSRC_VALIGN_TOP);
						} else {
							_notYourTurnIndicator.setVisible(true);
							txt.setResource(txt.createText("default-18.font", Color.white, playertext), RSRC_HALIGN_RIGHT|RSRC_VALIGN_TOP);
						}
					} else {
						if(player.hasTurn) {
							txt.setResource(txt.createText("default-18-bold.font", Color.white, playertext), RSRC_HALIGN_RIGHT|RSRC_VALIGN_TOP);
						} else {
							txt.setResource(txt.createText("default-18.font", Color.white, playertext), RSRC_HALIGN_RIGHT|RSRC_VALIGN_TOP);
						}
					}
				}
			}
		
		}
	}
	
	public static class GameLobby extends BScreen {

		private GameList list;
		private GameInfo[] _lastGameList;
		private BView players;
		public GameLobby(BApplication app) {
			super(app);
			init();
		}

		public GameLobby(BApplication app, Resolution theResolution) {
			super(app, theResolution);
			init();
		}

		public static class GameList extends BList {

		    public GameList(BView parent, int x, int y, int width, int height, int rowHeight) {
		    	super(parent, x, y, width, height, rowHeight);
		    	setBarAndArrows(BAR_DEFAULT, BAR_HANG, "quit", "join");
			}
		    
		    @Override
		    public boolean handleKeyPress(int code, long rawcode) {
		    	switch(code) {
		    	case KEY_SELECT:
		    		join(false);
		    		return true;
		    	}
		    	return super.handleKeyPress(code, rawcode);
		    }
		    
		    @Override
		    public boolean handleAction(BView view, Object action) {
		    	if(action.equals("join")) {
		    		return join(false);
		    	} else if(action.equals("quit")) {
		    		return quit();
		    	}
		    	return super.handleAction(view, action);
		    }

			private boolean quit() {
				getBApp().setActive(false);
				return true;
			}

			private boolean join(boolean spectate) {
				if(getFocus() < 0 || getFocus() >= size()) {
					return false;
				}
				GameInfo gameInfo = (GameInfo) get(getFocus());
				try {
					((Game) getBApp()).joinGame(gameInfo.gameID, spectate);
				} catch (Exception e) {
					return false;
				}
				return true;
			}

			@Override
			protected void createRow(BView parent, int index) {
				GameFactory factory = (GameFactory) getBApp().getFactory();
				if(get(index) instanceof GameInfo) {
					GameInfo gameInfo = (GameInfo) get(index);
					
					int playerCount = gameInfo.players.size();
					String players = playerListString(gameInfo, playerCount);

					String labelString;
					GameData game = factory.getGame().getGameData(gameInfo.gameID);
					if(playerCount <= 0) {
						labelString = "Start New Game";
					} else if(game != null && playerCount < game.minimumPlayers) {
						labelString = "Start Game with "+players;
					} else if(game == null || playerCount < game.maximumPlayers) {
						labelString = "Join Game with "+players;
					} else {
						labelString = "Watch Game between "+players;
					}
					
					parent.setResource(parent.createText("default-"+(getRowHeight()-2)+".font", Color.white, 
							"   "+labelString), RSRC_HALIGN_LEFT|RSRC_VALIGN_CENTER);
				}
			}

			private String playerListString(GameInfo gameInfo, int playerCount) {
				String players = "";
				if(playerCount > 0) {
					players += gameInfo.players.get(0).name;
					if(playerCount > 1) {
						if(playerCount == 2) {
							players += " and ";
						} else {
							players += ", ";
						}
						players += gameInfo.players.get(1).name;
						if(playerCount > 2) {
							players += ", ("+(playerCount-2)+" more)";
						}
					}
				}
				return players;
			}
		}
		
		protected void init() {
			Rectangle scrollRegion = ((Game)getBApp()).getScrollRegion();
			players = new BView(getNormal(), scrollRegion.x, scrollRegion.y, scrollRegion.width, 40);
			list = new GameList(getNormal(), scrollRegion.x, scrollRegion.y+players.getHeight(), scrollRegion.width, scrollRegion.height-players.getHeight(), 
					(int) (20*((Game)getBApp()).getHeightMultiplier()));
			setFocusDefault(list);
		}

		private void setupList() {
			GameFactory factory = (GameFactory) getBApp().getFactory();
			
			List<PlayerState> idle = factory.getIdlePlayers();
			idle.remove(((Game)getBApp())._player);
			String playerText;
			if(idle.size() == 0) {
				playerText = "When somebody is free to play, they will be listed in this space.";
			} else if(idle.size() == 1) {
				playerText = idle.get(0)+" is connected but not playing.";
			} else {
				playerText = idle.size() + " people are connected but not playing: "+ idle;
			}
			if(players.getResource() != null) {
				players.getResource().remove();
				players.clearResource();
			}
			players.setResource(players.createText("default-17-italic.font", Color.yellow, playerText));
			
			GameInfo[] gameList = factory.getGame().getGameList();
			
			if(_lastGameList != null && Arrays.equals( _lastGameList, gameList)) {
				return;
			}
			_lastGameList = gameList;
			
			int oldFocus = list.getFocus();
			if(oldFocus < 0) oldFocus = 0;
			list.clear();
//			this.bookmarkStartIndex = 3;//list.size();

			GameInfo newgame = new GameInfo();
			newgame.gameID = -1;
			newgame.players = new ArrayList<PlayerState>();
			list.add(newgame);
			
			if (gameList != null) {
				list.add(new Vector<GameInfo>(Arrays.asList(gameList)));
			}
			
			if(oldFocus >= list.size()) {
				oldFocus = list.size()-1;
			}
			list.setFocus(oldFocus, false);
		}
		
		@Override
		public boolean handleKeyPress(int code, long rawcode) {
			switch(code) {
			case KEY_CLEAR:
				getBApp().setActive(false);
				return true;
			}
			return super.handleKeyPress(code, rawcode);
		}
		
		@Override
		public boolean handleEnter(Object arg, boolean isReturn) {
			((Game)getBApp()).startLobbyUpdates();
			
			return super.handleEnter(arg, isReturn);
		}
		
		@Override
		public boolean handleExit() {
			((Game)getBApp()).stopLobbyUpdates();
			
			return super.handleExit();
		}
	}

	/**
	 * Instantiate the GameLobby instance appropriate to your game
	 */
	protected abstract GameLobby newGameLobby(Game app);

	/**
	 * instantiates and pushes the GameScreen, representing the joining of this
	 * user into the identified game.
	 * 
	 * @param gameID
	 *            the id of the game... negative values are flags for special
	 *            cases.
	 * @param spectate
	 *            true if the user should be joined as a spectator, not a player
	 */
	public abstract void joinGame(int gameID, boolean spectate);



}
